#define _GNU_SOURCE
#define ioctl ioctl_BARF
#define GL_GLEXT_PROTOTYPES 1

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>
#include <stdbool.h>
#include <setjmp.h>
#include <byteswap.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <ctype.h>
#include <inttypes.h>
#include <GLES2/gl2.h>
#include <GLES2/gl2ext.h>
#include <GLES3/gl3.h>
#include <EGL/egl.h>
#include <unistd.h>
#include <sys/syscall.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#define __packed __attribute__((packed))

#undef ioctl

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static void fatal_egl(const char *msg) {
  while (1) {
    EGLint error = eglGetError();
    if (error == EGL_SUCCESS)
      errx(1, "%s", msg);
    printf("[EGL error] 0x%x\n", error);
  }
}

struct drm_pvr_srvkm_cmd {
  uint32_t bridge_id;
  uint32_t bridge_func_id;
  uint64_t in_data_ptr;
  uint64_t out_data_ptr;
  uint32_t in_data_size;
  uint32_t out_data_size;
};
const char *bridge_name(uint32_t id) {
  if (id == 0) return "PVRSRV_BRIDGE_DEFAULT";
  if (id == 1) return "PVRSRV_BRIDGE_SRVCORE";
  if (id == 2) return "PVRSRV_BRIDGE_SYNC";
  if (id == 3) return "PVRSRV_BRIDGE_RESERVED1";
  if (id == 4) return "PVRSRV_BRIDGE_RESERVED2";
  if (id == 5) return "PVRSRV_BRIDGE_PDUMPCTRL";
  if (id == 6) return "PVRSRV_BRIDGE_MM";
  if (id == 7) return "PVRSRV_BRIDGE_MMPLAT";
  if (id == 128) return "PVRSRV_BRIDGE_RGXTQ";
  if (id == 130) return "PVRSRV_BRIDGE_RGXTA3D";
  return "PVRSRV_BRIDGE_<?>";
}

#define PVRSRV_BRIDGE_DEFAULT       0UL
#define PVRSRV_BRIDGE_SRVCORE       1UL
#define PVRSRV_BRIDGE_SYNC          2UL
#define PVRSRV_BRIDGE_RESERVED1       3UL
#define PVRSRV_BRIDGE_RESERVED2       4UL
#define PVRSRV_BRIDGE_PDUMPCTRL       5UL
#define PVRSRV_BRIDGE_MM          6UL
#define PVRSRV_BRIDGE_MMPLAT        7UL
#define PVRSRV_BRIDGE_CMM         8UL
#define PVRSRV_BRIDGE_PDUMPMM       9UL
#define PVRSRV_BRIDGE_PDUMP         10UL
#define PVRSRV_BRIDGE_DMABUF        11UL
#define PVRSRV_BRIDGE_DC          12UL
#define PVRSRV_BRIDGE_CACHE         13UL
#define PVRSRV_BRIDGE_SMM         14UL
#define PVRSRV_BRIDGE_PVRTL         15UL
#define PVRSRV_BRIDGE_RI          16UL
#define PVRSRV_BRIDGE_VALIDATION      17UL
#define PVRSRV_BRIDGE_TUTILS        18UL
#define PVRSRV_BRIDGE_DEVICEMEMHISTORY    19UL
#define PVRSRV_BRIDGE_HTBUFFER        20UL
#define PVRSRV_BRIDGE_DCPLAT        21UL
#define PVRSRV_BRIDGE_MMEXTMEM        22UL
#define PVRSRV_BRIDGE_SYNCTRACKING      23UL
#define PVRSRV_BRIDGE_SYNCFALLBACK      24UL
#define PVRSRV_BRIDGE_DI          25UL
#define PVRSRV_BRIDGE_DMA     26UL
#define PVRSRV_BRIDGE_RGXTQ                      128UL
#define PVRSRV_BRIDGE_RGXCMP                     129UL
#define PVRSRV_BRIDGE_RGXTA3D                    130UL
#define PVRSRV_BRIDGE_RGXBREAKPOINT                 131UL

#define PVRSRV_BRIDGE_SRVCORE_CMD_FIRST     0
#define PVRSRV_BRIDGE_SRVCORE_CONNECT     PVRSRV_BRIDGE_SRVCORE_CMD_FIRST+0
#define PVRSRV_BRIDGE_SRVCORE_DISCONNECT      PVRSRV_BRIDGE_SRVCORE_CMD_FIRST+1
#define PVRSRV_BRIDGE_SRVCORE_ACQUIREGLOBALEVENTOBJECT      PVRSRV_BRIDGE_SRVCORE_CMD_FIRST+2
#define PVRSRV_BRIDGE_SRVCORE_RELEASEGLOBALEVENTOBJECT      PVRSRV_BRIDGE_SRVCORE_CMD_FIRST+3
#define PVRSRV_BRIDGE_SRVCORE_EVENTOBJECTOPEN     PVRSRV_BRIDGE_SRVCORE_CMD_FIRST+4
#define PVRSRV_BRIDGE_SRVCORE_EVENTOBJECTWAIT     PVRSRV_BRIDGE_SRVCORE_CMD_FIRST+5
#define PVRSRV_BRIDGE_SRVCORE_EVENTOBJECTCLOSE      PVRSRV_BRIDGE_SRVCORE_CMD_FIRST+6
#define PVRSRV_BRIDGE_SRVCORE_DUMPDEBUGINFO     PVRSRV_BRIDGE_SRVCORE_CMD_FIRST+7
#define PVRSRV_BRIDGE_SRVCORE_GETDEVCLOCKSPEED      PVRSRV_BRIDGE_SRVCORE_CMD_FIRST+8
#define PVRSRV_BRIDGE_SRVCORE_HWOPTIMEOUT     PVRSRV_BRIDGE_SRVCORE_CMD_FIRST+9
#define PVRSRV_BRIDGE_SRVCORE_ALIGNMENTCHECK      PVRSRV_BRIDGE_SRVCORE_CMD_FIRST+10
#define PVRSRV_BRIDGE_SRVCORE_GETDEVICESTATUS     PVRSRV_BRIDGE_SRVCORE_CMD_FIRST+11
#define PVRSRV_BRIDGE_SRVCORE_GETMULTICOREINFO      PVRSRV_BRIDGE_SRVCORE_CMD_FIRST+12
#define PVRSRV_BRIDGE_SRVCORE_EVENTOBJECTWAITTIMEOUT      PVRSRV_BRIDGE_SRVCORE_CMD_FIRST+13
#define PVRSRV_BRIDGE_SRVCORE_FINDPROCESSMEMSTATS     PVRSRV_BRIDGE_SRVCORE_CMD_FIRST+14
#define PVRSRV_BRIDGE_SRVCORE_ACQUIREINFOPAGE     PVRSRV_BRIDGE_SRVCORE_CMD_FIRST+15
#define PVRSRV_BRIDGE_SRVCORE_RELEASEINFOPAGE     PVRSRV_BRIDGE_SRVCORE_CMD_FIRST+16

#define PVRSRV_BRIDGE_SYNC_CMD_FIRST      0
#define PVRSRV_BRIDGE_SYNC_ALLOCSYNCPRIMITIVEBLOCK      PVRSRV_BRIDGE_SYNC_CMD_FIRST+0
#define PVRSRV_BRIDGE_SYNC_FREESYNCPRIMITIVEBLOCK     PVRSRV_BRIDGE_SYNC_CMD_FIRST+1
#define PVRSRV_BRIDGE_SYNC_SYNCPRIMSET      PVRSRV_BRIDGE_SYNC_CMD_FIRST+2
#define PVRSRV_BRIDGE_SYNC_SYNCPRIMPDUMP      PVRSRV_BRIDGE_SYNC_CMD_FIRST+3
#define PVRSRV_BRIDGE_SYNC_SYNCPRIMPDUMPVALUE     PVRSRV_BRIDGE_SYNC_CMD_FIRST+4
#define PVRSRV_BRIDGE_SYNC_SYNCPRIMPDUMPPOL     PVRSRV_BRIDGE_SYNC_CMD_FIRST+5
#define PVRSRV_BRIDGE_SYNC_SYNCPRIMPDUMPCBP     PVRSRV_BRIDGE_SYNC_CMD_FIRST+6
#define PVRSRV_BRIDGE_SYNC_SYNCALLOCEVENT     PVRSRV_BRIDGE_SYNC_CMD_FIRST+7
#define PVRSRV_BRIDGE_SYNC_SYNCFREEEVENT      PVRSRV_BRIDGE_SYNC_CMD_FIRST+8
#define PVRSRV_BRIDGE_SYNC_SYNCCHECKPOINTSIGNALLEDPDUMPPOL      PVRSRV_BRIDGE_SYNC_CMD_FIRST+9

#define PVRSRV_BRIDGE_MM_CMD_FIRST      0
#define PVRSRV_BRIDGE_MM_PMREXPORTPMR     PVRSRV_BRIDGE_MM_CMD_FIRST+0
#define PVRSRV_BRIDGE_MM_PMRUNEXPORTPMR     PVRSRV_BRIDGE_MM_CMD_FIRST+1
#define PVRSRV_BRIDGE_MM_PMRGETUID      PVRSRV_BRIDGE_MM_CMD_FIRST+2
#define PVRSRV_BRIDGE_MM_PMRMAKELOCALIMPORTHANDLE     PVRSRV_BRIDGE_MM_CMD_FIRST+3
#define PVRSRV_BRIDGE_MM_PMRUNMAKELOCALIMPORTHANDLE     PVRSRV_BRIDGE_MM_CMD_FIRST+4
#define PVRSRV_BRIDGE_MM_PMRIMPORTPMR     PVRSRV_BRIDGE_MM_CMD_FIRST+5
#define PVRSRV_BRIDGE_MM_PMRLOCALIMPORTPMR      PVRSRV_BRIDGE_MM_CMD_FIRST+6
#define PVRSRV_BRIDGE_MM_PMRUNREFPMR      PVRSRV_BRIDGE_MM_CMD_FIRST+7
#define PVRSRV_BRIDGE_MM_PMRUNREFUNLOCKPMR      PVRSRV_BRIDGE_MM_CMD_FIRST+8
#define PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDPMR     PVRSRV_BRIDGE_MM_CMD_FIRST+9
#define PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR     PVRSRV_BRIDGE_MM_CMD_FIRST+10
#define PVRSRV_BRIDGE_MM_DEVMEMINTPIN     PVRSRV_BRIDGE_MM_CMD_FIRST+11
#define PVRSRV_BRIDGE_MM_DEVMEMINTUNPIN     PVRSRV_BRIDGE_MM_CMD_FIRST+12
#define PVRSRV_BRIDGE_MM_DEVMEMINTPINVALIDATE     PVRSRV_BRIDGE_MM_CMD_FIRST+13
#define PVRSRV_BRIDGE_MM_DEVMEMINTUNPININVALIDATE     PVRSRV_BRIDGE_MM_CMD_FIRST+14
#define PVRSRV_BRIDGE_MM_DEVMEMINTCTXCREATE     PVRSRV_BRIDGE_MM_CMD_FIRST+15
#define PVRSRV_BRIDGE_MM_DEVMEMINTCTXDESTROY      PVRSRV_BRIDGE_MM_CMD_FIRST+16
#define PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE      PVRSRV_BRIDGE_MM_CMD_FIRST+17
#define PVRSRV_BRIDGE_MM_DEVMEMINTHEAPDESTROY     PVRSRV_BRIDGE_MM_CMD_FIRST+18
#define PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR      PVRSRV_BRIDGE_MM_CMD_FIRST+19
#define PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPMR      PVRSRV_BRIDGE_MM_CMD_FIRST+20
#define PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE      PVRSRV_BRIDGE_MM_CMD_FIRST+21
#define PVRSRV_BRIDGE_MM_DEVMEMINTUNRESERVERANGE      PVRSRV_BRIDGE_MM_CMD_FIRST+22
#define PVRSRV_BRIDGE_MM_CHANGESPARSEMEM      PVRSRV_BRIDGE_MM_CMD_FIRST+23
#define PVRSRV_BRIDGE_MM_DEVMEMINTMAPPAGES      PVRSRV_BRIDGE_MM_CMD_FIRST+24
#define PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPAGES      PVRSRV_BRIDGE_MM_CMD_FIRST+25
#define PVRSRV_BRIDGE_MM_DEVMEMISVDEVADDRVALID      PVRSRV_BRIDGE_MM_CMD_FIRST+26
#define PVRSRV_BRIDGE_MM_DEVMEMFLUSHDEVSLCRANGE     PVRSRV_BRIDGE_MM_CMD_FIRST+27
#define PVRSRV_BRIDGE_MM_DEVMEMINVALIDATEFBSCTABLE      PVRSRV_BRIDGE_MM_CMD_FIRST+28
#define PVRSRV_BRIDGE_MM_HEAPCFGHEAPCONFIGCOUNT     PVRSRV_BRIDGE_MM_CMD_FIRST+29
#define PVRSRV_BRIDGE_MM_HEAPCFGHEAPCOUNT     PVRSRV_BRIDGE_MM_CMD_FIRST+30
#define PVRSRV_BRIDGE_MM_HEAPCFGHEAPCONFIGNAME      PVRSRV_BRIDGE_MM_CMD_FIRST+31
#define PVRSRV_BRIDGE_MM_HEAPCFGHEAPDETAILS     PVRSRV_BRIDGE_MM_CMD_FIRST+32
#define PVRSRV_BRIDGE_MM_DEVMEMINTREGISTERPFNOTIFYKM      PVRSRV_BRIDGE_MM_CMD_FIRST+33
#define PVRSRV_BRIDGE_MM_GETMAXPHYSHEAPCOUNT      PVRSRV_BRIDGE_MM_CMD_FIRST+34
#define PVRSRV_BRIDGE_MM_PHYSHEAPGETMEMINFO     PVRSRV_BRIDGE_MM_CMD_FIRST+35
#define PVRSRV_BRIDGE_MM_GETDEFAULTPHYSICALHEAP     PVRSRV_BRIDGE_MM_CMD_FIRST+36
#define PVRSRV_BRIDGE_MM_GETHEAPPHYSMEMUSAGE      PVRSRV_BRIDGE_MM_CMD_FIRST+37
#define PVRSRV_BRIDGE_MM_DEVMEMGETFAULTADDRESS      PVRSRV_BRIDGE_MM_CMD_FIRST+38
#define PVRSRV_BRIDGE_MM_PVRSRVUPDATEOOMSTATS     PVRSRV_BRIDGE_MM_CMD_FIRST+39
#define PVRSRV_BRIDGE_MM_PHYSHEAPGETMEMINFOPKD      PVRSRV_BRIDGE_MM_CMD_FIRST+40
#define PVRSRV_BRIDGE_MM_GETHEAPPHYSMEMUSAGEPKD     PVRSRV_BRIDGE_MM_CMD_FIRST+41
#define PVRSRV_BRIDGE_MM_DEVMEMXINTRESERVERANGE     PVRSRV_BRIDGE_MM_CMD_FIRST+42
#define PVRSRV_BRIDGE_MM_DEVMEMXINTUNRESERVERANGE     PVRSRV_BRIDGE_MM_CMD_FIRST+43
#define PVRSRV_BRIDGE_MM_DEVMEMXINTMAPPAGES     PVRSRV_BRIDGE_MM_CMD_FIRST+44
#define PVRSRV_BRIDGE_MM_DEVMEMXINTUNMAPPAGES     PVRSRV_BRIDGE_MM_CMD_FIRST+45
#define PVRSRV_BRIDGE_MM_CHANGESPARSEMEM2     PVRSRV_BRIDGE_MM_CMD_FIRST+46
#define PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE2     PVRSRV_BRIDGE_MM_CMD_FIRST+47
#define PVRSRV_BRIDGE_MM_DEVMEMINTUNRESERVERANGE2     PVRSRV_BRIDGE_MM_CMD_FIRST+48
#define PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR2     PVRSRV_BRIDGE_MM_CMD_FIRST+49
#define PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPMR2     PVRSRV_BRIDGE_MM_CMD_FIRST+50
#define PVRSRV_BRIDGE_MM_CMD_LAST     (PVRSRV_BRIDGE_MM_CMD_FIRST+50)

#define PVRSRV_BRIDGE_RGXTA3D_CMD_FIRST     0
#define PVRSRV_BRIDGE_RGXTA3D_RGXCREATEHWRTDATASET      PVRSRV_BRIDGE_RGXTA3D_CMD_FIRST+0
#define PVRSRV_BRIDGE_RGXTA3D_RGXDESTROYHWRTDATASET     PVRSRV_BRIDGE_RGXTA3D_CMD_FIRST+1
#define PVRSRV_BRIDGE_RGXTA3D_RGXCREATEZSBUFFER     PVRSRV_BRIDGE_RGXTA3D_CMD_FIRST+2
#define PVRSRV_BRIDGE_RGXTA3D_RGXDESTROYZSBUFFER      PVRSRV_BRIDGE_RGXTA3D_CMD_FIRST+3
#define PVRSRV_BRIDGE_RGXTA3D_RGXPOPULATEZSBUFFER     PVRSRV_BRIDGE_RGXTA3D_CMD_FIRST+4
#define PVRSRV_BRIDGE_RGXTA3D_RGXUNPOPULATEZSBUFFER     PVRSRV_BRIDGE_RGXTA3D_CMD_FIRST+5
#define PVRSRV_BRIDGE_RGXTA3D_RGXCREATEFREELIST     PVRSRV_BRIDGE_RGXTA3D_CMD_FIRST+6
#define PVRSRV_BRIDGE_RGXTA3D_RGXDESTROYFREELIST      PVRSRV_BRIDGE_RGXTA3D_CMD_FIRST+7
#define PVRSRV_BRIDGE_RGXTA3D_RGXCREATERENDERCONTEXT      PVRSRV_BRIDGE_RGXTA3D_CMD_FIRST+8
#define PVRSRV_BRIDGE_RGXTA3D_RGXDESTROYRENDERCONTEXT     PVRSRV_BRIDGE_RGXTA3D_CMD_FIRST+9
#define PVRSRV_BRIDGE_RGXTA3D_RGXSETRENDERCONTEXTPRIORITY     PVRSRV_BRIDGE_RGXTA3D_CMD_FIRST+10
#define PVRSRV_BRIDGE_RGXTA3D_RGXRENDERCONTEXTSTALLED     PVRSRV_BRIDGE_RGXTA3D_CMD_FIRST+11
#define PVRSRV_BRIDGE_RGXTA3D_RGXKICKTA3D2      PVRSRV_BRIDGE_RGXTA3D_CMD_FIRST+12
#define PVRSRV_BRIDGE_RGXTA3D_RGXSETRENDERCONTEXTPROPERTY     PVRSRV_BRIDGE_RGXTA3D_CMD_FIRST+13
#define PVRSRV_BRIDGE_RGXTA3D_RGXCREATEZSBUFFER2      PVRSRV_BRIDGE_RGXTA3D_CMD_FIRST+14
#define PVRSRV_BRIDGE_RGXTA3D_RGXCREATEFREELIST2      PVRSRV_BRIDGE_RGXTA3D_CMD_FIRST+15

const char *bridge_func_name(uint32_t bridge, uint32_t func) {
  if (bridge == PVRSRV_BRIDGE_SRVCORE) {
    if (func == PVRSRV_BRIDGE_SRVCORE_CONNECT) return "PVRSRV_BRIDGE_SRVCORE_CONNECT";
    if (func == PVRSRV_BRIDGE_SRVCORE_DISCONNECT) return "PVRSRV_BRIDGE_SRVCORE_DISCONNECT";
    if (func == PVRSRV_BRIDGE_SRVCORE_ACQUIREGLOBALEVENTOBJECT) return "PVRSRV_BRIDGE_SRVCORE_ACQUIREGLOBALEVENTOBJECT";
    if (func == PVRSRV_BRIDGE_SRVCORE_RELEASEGLOBALEVENTOBJECT) return "PVRSRV_BRIDGE_SRVCORE_RELEASEGLOBALEVENTOBJECT";
    if (func == PVRSRV_BRIDGE_SRVCORE_EVENTOBJECTOPEN) return "PVRSRV_BRIDGE_SRVCORE_EVENTOBJECTOPEN";
    if (func == PVRSRV_BRIDGE_SRVCORE_EVENTOBJECTWAIT) return "PVRSRV_BRIDGE_SRVCORE_EVENTOBJECTWAIT";
    if (func == PVRSRV_BRIDGE_SRVCORE_EVENTOBJECTCLOSE) return "PVRSRV_BRIDGE_SRVCORE_EVENTOBJECTCLOSE";
    if (func == PVRSRV_BRIDGE_SRVCORE_DUMPDEBUGINFO) return "PVRSRV_BRIDGE_SRVCORE_DUMPDEBUGINFO";
    if (func == PVRSRV_BRIDGE_SRVCORE_GETDEVCLOCKSPEED) return "PVRSRV_BRIDGE_SRVCORE_GETDEVCLOCKSPEED";
    if (func == PVRSRV_BRIDGE_SRVCORE_HWOPTIMEOUT) return "PVRSRV_BRIDGE_SRVCORE_HWOPTIMEOUT";
    if (func == PVRSRV_BRIDGE_SRVCORE_ALIGNMENTCHECK) return "PVRSRV_BRIDGE_SRVCORE_ALIGNMENTCHECK";
    if (func == PVRSRV_BRIDGE_SRVCORE_GETDEVICESTATUS) return "PVRSRV_BRIDGE_SRVCORE_GETDEVICESTATUS";
    if (func == PVRSRV_BRIDGE_SRVCORE_GETMULTICOREINFO) return "PVRSRV_BRIDGE_SRVCORE_GETMULTICOREINFO";
    if (func == PVRSRV_BRIDGE_SRVCORE_EVENTOBJECTWAITTIMEOUT) return "PVRSRV_BRIDGE_SRVCORE_EVENTOBJECTWAITTIMEOUT";
    if (func == PVRSRV_BRIDGE_SRVCORE_FINDPROCESSMEMSTATS) return "PVRSRV_BRIDGE_SRVCORE_FINDPROCESSMEMSTATS";
    if (func == PVRSRV_BRIDGE_SRVCORE_ACQUIREINFOPAGE) return "PVRSRV_BRIDGE_SRVCORE_ACQUIREINFOPAGE";
    if (func == PVRSRV_BRIDGE_SRVCORE_RELEASEINFOPAGE) return "PVRSRV_BRIDGE_SRVCORE_RELEASEINFOPAGE";
  }
  if (bridge == PVRSRV_BRIDGE_SYNC) {
    if (func == PVRSRV_BRIDGE_SYNC_ALLOCSYNCPRIMITIVEBLOCK) return "PVRSRV_BRIDGE_SYNC_ALLOCSYNCPRIMITIVEBLOCK";
    if (func == PVRSRV_BRIDGE_SYNC_FREESYNCPRIMITIVEBLOCK) return "PVRSRV_BRIDGE_SYNC_FREESYNCPRIMITIVEBLOCK";
    if (func == PVRSRV_BRIDGE_SYNC_SYNCPRIMSET) return "PVRSRV_BRIDGE_SYNC_SYNCPRIMSET";
    if (func == PVRSRV_BRIDGE_SYNC_SYNCPRIMPDUMP) return "PVRSRV_BRIDGE_SYNC_SYNCPRIMPDUMP";
    if (func == PVRSRV_BRIDGE_SYNC_SYNCPRIMPDUMPVALUE) return "PVRSRV_BRIDGE_SYNC_SYNCPRIMPDUMPVALUE";
    if (func == PVRSRV_BRIDGE_SYNC_SYNCPRIMPDUMPPOL) return "PVRSRV_BRIDGE_SYNC_SYNCPRIMPDUMPPOL";
    if (func == PVRSRV_BRIDGE_SYNC_SYNCPRIMPDUMPCBP) return "PVRSRV_BRIDGE_SYNC_SYNCPRIMPDUMPCBP";
    if (func == PVRSRV_BRIDGE_SYNC_SYNCALLOCEVENT) return "PVRSRV_BRIDGE_SYNC_SYNCALLOCEVENT";
    if (func == PVRSRV_BRIDGE_SYNC_SYNCFREEEVENT) return "PVRSRV_BRIDGE_SYNC_SYNCFREEEVENT";
    if (func == PVRSRV_BRIDGE_SYNC_SYNCCHECKPOINTSIGNALLEDPDUMPPOL) return "PVRSRV_BRIDGE_SYNC_SYNCCHECKPOINTSIGNALLEDPDUMPPOL";
  }
  if (bridge == PVRSRV_BRIDGE_MM) {
    if (func == PVRSRV_BRIDGE_MM_PMREXPORTPMR) return "PVRSRV_BRIDGE_MM_PMREXPORTPMR";
    if (func == PVRSRV_BRIDGE_MM_PMRUNEXPORTPMR) return "PVRSRV_BRIDGE_MM_PMRUNEXPORTPMR";
    if (func == PVRSRV_BRIDGE_MM_PMRGETUID) return "PVRSRV_BRIDGE_MM_PMRGETUID";
    if (func == PVRSRV_BRIDGE_MM_PMRMAKELOCALIMPORTHANDLE) return "PVRSRV_BRIDGE_MM_PMRMAKELOCALIMPORTHANDLE";
    if (func == PVRSRV_BRIDGE_MM_PMRUNMAKELOCALIMPORTHANDLE) return "PVRSRV_BRIDGE_MM_PMRUNMAKELOCALIMPORTHANDLE";
    if (func == PVRSRV_BRIDGE_MM_PMRIMPORTPMR) return "PVRSRV_BRIDGE_MM_PMRIMPORTPMR";
    if (func == PVRSRV_BRIDGE_MM_PMRLOCALIMPORTPMR) return "PVRSRV_BRIDGE_MM_PMRLOCALIMPORTPMR";
    if (func == PVRSRV_BRIDGE_MM_PMRUNREFPMR) return "PVRSRV_BRIDGE_MM_PMRUNREFPMR";
    if (func == PVRSRV_BRIDGE_MM_PMRUNREFUNLOCKPMR) return "PVRSRV_BRIDGE_MM_PMRUNREFUNLOCKPMR";
    if (func == PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDPMR) return "PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDPMR";
    if (func == PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR) return "PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTPIN) return "PVRSRV_BRIDGE_MM_DEVMEMINTPIN";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTUNPIN) return "PVRSRV_BRIDGE_MM_DEVMEMINTUNPIN";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTPINVALIDATE) return "PVRSRV_BRIDGE_MM_DEVMEMINTPINVALIDATE";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTUNPININVALIDATE) return "PVRSRV_BRIDGE_MM_DEVMEMINTUNPININVALIDATE";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTCTXCREATE) return "PVRSRV_BRIDGE_MM_DEVMEMINTCTXCREATE";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTCTXDESTROY) return "PVRSRV_BRIDGE_MM_DEVMEMINTCTXDESTROY";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE) return "PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTHEAPDESTROY) return "PVRSRV_BRIDGE_MM_DEVMEMINTHEAPDESTROY";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR) return "PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPMR) return "PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPMR";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE) return "PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTUNRESERVERANGE) return "PVRSRV_BRIDGE_MM_DEVMEMINTUNRESERVERANGE";
    if (func == PVRSRV_BRIDGE_MM_CHANGESPARSEMEM) return "PVRSRV_BRIDGE_MM_CHANGESPARSEMEM";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTMAPPAGES) return "PVRSRV_BRIDGE_MM_DEVMEMINTMAPPAGES";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPAGES) return "PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPAGES";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMISVDEVADDRVALID) return "PVRSRV_BRIDGE_MM_DEVMEMISVDEVADDRVALID";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMFLUSHDEVSLCRANGE) return "PVRSRV_BRIDGE_MM_DEVMEMFLUSHDEVSLCRANGE";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINVALIDATEFBSCTABLE) return "PVRSRV_BRIDGE_MM_DEVMEMINVALIDATEFBSCTABLE";
    if (func == PVRSRV_BRIDGE_MM_HEAPCFGHEAPCONFIGCOUNT) return "PVRSRV_BRIDGE_MM_HEAPCFGHEAPCONFIGCOUNT";
    if (func == PVRSRV_BRIDGE_MM_HEAPCFGHEAPCOUNT) return "PVRSRV_BRIDGE_MM_HEAPCFGHEAPCOUNT";
    if (func == PVRSRV_BRIDGE_MM_HEAPCFGHEAPCONFIGNAME) return "PVRSRV_BRIDGE_MM_HEAPCFGHEAPCONFIGNAME";
    if (func == PVRSRV_BRIDGE_MM_HEAPCFGHEAPDETAILS) return "PVRSRV_BRIDGE_MM_HEAPCFGHEAPDETAILS";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTREGISTERPFNOTIFYKM) return "PVRSRV_BRIDGE_MM_DEVMEMINTREGISTERPFNOTIFYKM";
    if (func == PVRSRV_BRIDGE_MM_GETMAXPHYSHEAPCOUNT) return "PVRSRV_BRIDGE_MM_GETMAXPHYSHEAPCOUNT";
    if (func == PVRSRV_BRIDGE_MM_PHYSHEAPGETMEMINFO) return "PVRSRV_BRIDGE_MM_PHYSHEAPGETMEMINFO";
    if (func == PVRSRV_BRIDGE_MM_GETDEFAULTPHYSICALHEAP) return "PVRSRV_BRIDGE_MM_GETDEFAULTPHYSICALHEAP";
    if (func == PVRSRV_BRIDGE_MM_GETHEAPPHYSMEMUSAGE) return "PVRSRV_BRIDGE_MM_GETHEAPPHYSMEMUSAGE";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMGETFAULTADDRESS) return "PVRSRV_BRIDGE_MM_DEVMEMGETFAULTADDRESS";
    if (func == PVRSRV_BRIDGE_MM_PVRSRVUPDATEOOMSTATS) return "PVRSRV_BRIDGE_MM_PVRSRVUPDATEOOMSTATS";
    if (func == PVRSRV_BRIDGE_MM_PHYSHEAPGETMEMINFOPKD) return "PVRSRV_BRIDGE_MM_PHYSHEAPGETMEMINFOPKD";
    if (func == PVRSRV_BRIDGE_MM_GETHEAPPHYSMEMUSAGEPKD) return "PVRSRV_BRIDGE_MM_GETHEAPPHYSMEMUSAGEPKD";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMXINTRESERVERANGE) return "PVRSRV_BRIDGE_MM_DEVMEMXINTRESERVERANGE";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMXINTUNRESERVERANGE) return "PVRSRV_BRIDGE_MM_DEVMEMXINTUNRESERVERANGE";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMXINTMAPPAGES) return "PVRSRV_BRIDGE_MM_DEVMEMXINTMAPPAGES";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMXINTUNMAPPAGES) return "PVRSRV_BRIDGE_MM_DEVMEMXINTUNMAPPAGES";
    if (func == PVRSRV_BRIDGE_MM_CHANGESPARSEMEM2) return "PVRSRV_BRIDGE_MM_CHANGESPARSEMEM2";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE2) return "PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE2";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTUNRESERVERANGE2) return "PVRSRV_BRIDGE_MM_DEVMEMINTUNRESERVERANGE2";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR2) return "PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR2";
    if (func == PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPMR2) return "PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPMR2";
  }
  if (bridge == PVRSRV_BRIDGE_RGXTA3D) {
    if (func == PVRSRV_BRIDGE_RGXTA3D_RGXCREATEHWRTDATASET) return "PVRSRV_BRIDGE_RGXTA3D_RGXCREATEHWRTDATASET";
    if (func == PVRSRV_BRIDGE_RGXTA3D_RGXDESTROYHWRTDATASET) return "PVRSRV_BRIDGE_RGXTA3D_RGXDESTROYHWRTDATASET";
    if (func == PVRSRV_BRIDGE_RGXTA3D_RGXCREATEZSBUFFER) return "PVRSRV_BRIDGE_RGXTA3D_RGXCREATEZSBUFFER";
    if (func == PVRSRV_BRIDGE_RGXTA3D_RGXDESTROYZSBUFFER) return "PVRSRV_BRIDGE_RGXTA3D_RGXDESTROYZSBUFFER";
    if (func == PVRSRV_BRIDGE_RGXTA3D_RGXPOPULATEZSBUFFER) return "PVRSRV_BRIDGE_RGXTA3D_RGXPOPULATEZSBUFFER";
    if (func == PVRSRV_BRIDGE_RGXTA3D_RGXUNPOPULATEZSBUFFER) return "PVRSRV_BRIDGE_RGXTA3D_RGXUNPOPULATEZSBUFFER";
    if (func == PVRSRV_BRIDGE_RGXTA3D_RGXCREATEFREELIST) return "PVRSRV_BRIDGE_RGXTA3D_RGXCREATEFREELIST";
    if (func == PVRSRV_BRIDGE_RGXTA3D_RGXDESTROYFREELIST) return "PVRSRV_BRIDGE_RGXTA3D_RGXDESTROYFREELIST";
    if (func == PVRSRV_BRIDGE_RGXTA3D_RGXCREATERENDERCONTEXT) return "PVRSRV_BRIDGE_RGXTA3D_RGXCREATERENDERCONTEXT";
    if (func == PVRSRV_BRIDGE_RGXTA3D_RGXDESTROYRENDERCONTEXT) return "PVRSRV_BRIDGE_RGXTA3D_RGXDESTROYRENDERCONTEXT";
    if (func == PVRSRV_BRIDGE_RGXTA3D_RGXSETRENDERCONTEXTPRIORITY) return "PVRSRV_BRIDGE_RGXTA3D_RGXSETRENDERCONTEXTPRIORITY";
    if (func == PVRSRV_BRIDGE_RGXTA3D_RGXRENDERCONTEXTSTALLED) return "PVRSRV_BRIDGE_RGXTA3D_RGXRENDERCONTEXTSTALLED";
    if (func == PVRSRV_BRIDGE_RGXTA3D_RGXKICKTA3D2) return "PVRSRV_BRIDGE_RGXTA3D_RGXKICKTA3D2";
    if (func == PVRSRV_BRIDGE_RGXTA3D_RGXSETRENDERCONTEXTPROPERTY) return "PVRSRV_BRIDGE_RGXTA3D_RGXSETRENDERCONTEXTPROPERTY";
    if (func == PVRSRV_BRIDGE_RGXTA3D_RGXCREATEZSBUFFER2) return "PVRSRV_BRIDGE_RGXTA3D_RGXCREATEZSBUFFER2";
    if (func == PVRSRV_BRIDGE_RGXTA3D_RGXCREATEFREELIST2) return "PVRSRV_BRIDGE_RGXTA3D_RGXCREATEFREELIST2";
  }
  return "<?>";
}

typedef uint64_t IMG_DEV_VIRTADDR;
#define IMG_UINT64 uint64_t
#define IMG_UINT32 uint32_t
#define IMG_CHAR char
#define IMG_UINT8 uint8_t
#define IMG_HANDLE uint64_t
#define IMG_BYTE uint8_t
typedef int32_t PVRSRV_FENCE;
typedef int32_t PVRSRV_TIMELINE;
typedef int PVRSRV_ERROR;
typedef IMG_UINT64 PVRSRV_MEMALLOCFLAGS_T;
typedef uint64_t IMG_DEVMEM_SIZE_T;
typedef uint64_t IMG_DEVMEM_ALIGN_T;
typedef uint64_t IMG_DEVMEM_OFFSET_T;
typedef uint32_t IMG_DEVMEM_LOG2ALIGN_T;
typedef enum tag_img_bool
{
  IMG_FALSE   = 0,
  IMG_TRUE    = 1,
  IMG_FORCE_ALIGN = 0x7FFFFFFF
} IMG_BOOL, *IMG_PBOOL;

typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2_TAG
{
  IMG_DEV_VIRTADDR sAddress;
  IMG_DEVMEM_SIZE_T uiLength;
  IMG_HANDLE hDevmemServerHeap;
  PVRSRV_MEMALLOCFLAGS_T uiFlags;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2_TAG
{
  IMG_HANDLE hReservation;
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2;

typedef struct PVRSRV_BRIDGE_IN_RGXCREATETRANSFERCONTEXT_TAG
{
  IMG_UINT64 ui64RobustnessAddress;
  IMG_HANDLE hPrivData;
  uint64_t/* IMG_BYTE* */pui8FrameworkCmd;
  IMG_UINT32 ui32ContextFlags;
  IMG_UINT32 ui32FrameworkCmdize;
  IMG_UINT32 ui32PackedCCBSizeU8888;
  IMG_UINT32 ui32Priority;
} __packed PVRSRV_BRIDGE_IN_RGXCREATETRANSFERCONTEXT;
typedef struct PVRSRV_BRIDGE_OUT_RGXCREATETRANSFERCONTEXT_TAG
{
  IMG_HANDLE hCLIPMRMem;
  IMG_HANDLE hTransferContext;
  IMG_HANDLE hUSCPMRMem;
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_RGXCREATETRANSFERCONTEXT;

typedef struct PVRSRV_BRIDGE_IN_RGXSUBMITTRANSFER2_TAG
{
  IMG_HANDLE hTransferContext;
  uint64_t/* IMG_UINT32* */pui32ClientUpdateCount;
  uint64_t/* IMG_UINT32* */pui32CommandSize;
  uint64_t/* IMG_UINT32* */pui32SyncPMRFlags;
  uint64_t/* IMG_UINT32* */pui32TQPrepareFlags;
  uint64_t/* IMG_UINT32** */pui32UpdateSyncOffset;
  uint64_t/* IMG_UINT32** */pui32UpdateValue;
  uint64_t/* IMG_UINT8** */pui8FWCommand;
  uint64_t/* IMG_CHAR* */puiUpdateFenceName;
  uint64_t/* IMG_HANDLE* */phSyncPMRs;
  uint64_t/* IMG_HANDLE** */phUpdateUFOSyncPrimBlock;
  PVRSRV_TIMELINE h2DUpdateTimeline;
  PVRSRV_TIMELINE h3DUpdateTimeline;
  PVRSRV_FENCE hCheckFenceFD;
  IMG_UINT32 ui32ExtJobRef;
  IMG_UINT32 ui32PrepareCount;
  IMG_UINT32 ui32SyncPMRCount;
} __packed PVRSRV_BRIDGE_IN_RGXSUBMITTRANSFER2;
typedef struct PVRSRV_BRIDGE_OUT_RGXSUBMITTRANSFER2_TAG
{
  PVRSRV_ERROR eError;
  PVRSRV_FENCE h2DUpdateFence;
  PVRSRV_FENCE h3DUpdateFence;
} __packed PVRSRV_BRIDGE_OUT_RGXSUBMITTRANSFER2;

typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2_TAG
{
  IMG_HANDLE hDevmemServerHeap;
  IMG_HANDLE hPMR;
  IMG_HANDLE hReservation;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2_TAG
{
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2;

typedef struct PVRSRV_BRIDGE_IN_PMRLOCALIMPORTPMR_TAG
{
  IMG_HANDLE hExtHandle;
} __packed PVRSRV_BRIDGE_IN_PMRLOCALIMPORTPMR;
typedef struct PVRSRV_BRIDGE_OUT_PMRLOCALIMPORTPMR_TAG
{
  IMG_DEVMEM_ALIGN_T uiAlign;
  IMG_DEVMEM_SIZE_T uiSize;
  IMG_HANDLE hPMR;
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_PMRLOCALIMPORTPMR;

typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTUNRESERVERANGE2_TAG
{
  IMG_HANDLE hReservation;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTUNRESERVERANGE2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTUNRESERVERANGE2_TAG
{
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTUNRESERVERANGE2;

#define RGX_TQ3DPARAMETERS_HEAP_BASE        0xE400000000
#define RGX_TQ3DPARAMETERS_HEAP_SIZE        0x0400000000

typedef struct PVRSRV_BRIDGE_IN_RGXPOPULATEZSBUFFER_TAG
{
  IMG_HANDLE hsZSBufferKM;
} __packed PVRSRV_BRIDGE_IN_RGXPOPULATEZSBUFFER;
typedef struct PVRSRV_BRIDGE_OUT_RGXPOPULATEZSBUFFER_TAG
{
  IMG_HANDLE hsPopulation;
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_RGXPOPULATEZSBUFFER;

typedef struct PVRSRV_BRIDGE_IN_RGXCREATEZSBUFFER2_TAG
{
  IMG_HANDLE hPMR;
  IMG_HANDLE hReservation;
  PVRSRV_MEMALLOCFLAGS_T uiMapFlags;
} __packed PVRSRV_BRIDGE_IN_RGXCREATEZSBUFFER2;
typedef struct PVRSRV_BRIDGE_OUT_RGXCREATEZSBUFFER2_TAG
{
  IMG_HANDLE hsZSBufferKM;
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_RGXCREATEZSBUFFER2;

typedef struct PVRSRV_BRIDGE_IN_RGXKICKTA3D2_TAG
{
  IMG_UINT64 ui64Deadline;
  IMG_HANDLE hKMHWRTDataSet;
  IMG_HANDLE hMSAAScratchBuffer;
  IMG_HANDLE hPRFenceUFOSyncPrimBlock;
  IMG_HANDLE hRenderContext;
  IMG_HANDLE hZSBuffer;
  uint64_t /* IMG_UINT32* */ pui32Client3DUpdateSyncOffset;
  uint64_t /* IMG_UINT32* */ pui32Client3DUpdateValue;
  uint64_t /* IMG_UINT32* */ pui32ClientTAFenceSyncOffset;
  uint64_t /* IMG_UINT32* */ pui32ClientTAFenceValue;
  uint64_t /* IMG_UINT32* */ pui32ClientTAUpdateSyncOffset;
  uint64_t /* IMG_UINT32* */ pui32ClientTAUpdateValue;
  uint64_t /* IMG_UINT32* */ pui32SyncPMRFlags;
  uint64_t /* IMG_BYTE* */ pui83DCmd;
  uint64_t /* IMG_BYTE* */ pui83DPRCmd;
  uint64_t /* IMG_BYTE* */ pui8TACmd;
  uint64_t /* IMG_CHAR* */ puiUpdateFenceName;
  uint64_t /* IMG_CHAR* */ puiUpdateFenceName3D;
  uint64_t /* IMG_HANDLE* */ phClient3DUpdateSyncPrimBlock;
  uint64_t /* IMG_HANDLE* */ phClientTAFenceSyncPrimBlock;
  uint64_t /* IMG_HANDLE* */ phClientTAUpdateSyncPrimBlock;
  uint64_t /* IMG_HANDLE* */ phSyncPMRs;
  IMG_BOOL bbAbort;
  IMG_BOOL bbKick3D;
  IMG_BOOL bbKickPR;
  IMG_BOOL bbKickTA;
  PVRSRV_FENCE hCheckFence;
  PVRSRV_FENCE hCheckFence3D;
  PVRSRV_TIMELINE hUpdateTimeline;
  PVRSRV_TIMELINE hUpdateTimeline3D;
  IMG_UINT32 ui323DCmdSize;
  IMG_UINT32 ui323DPRCmdSize;
  IMG_UINT32 ui32Client3DUpdateCount;
  IMG_UINT32 ui32ClientTAFenceCount;
  IMG_UINT32 ui32ClientTAUpdateCount;
  IMG_UINT32 ui32ExtJobRef;
  IMG_UINT32 ui32NumberOfDrawCalls;
  IMG_UINT32 ui32NumberOfIndices;
  IMG_UINT32 ui32NumberOfMRTs;
  IMG_UINT32 ui32PDumpFlags;
  IMG_UINT32 ui32PRFenceUFOSyncOffset;
  IMG_UINT32 ui32PRFenceValue;
  IMG_UINT32 ui32RenderTargetSize;
  IMG_UINT32 ui32SyncPMRCount;
  IMG_UINT32 ui32TACmdSize;
} __packed PVRSRV_BRIDGE_IN_RGXKICKTA3D2;
typedef struct PVRSRV_BRIDGE_OUT_RGXKICKTA3D2_TAG
{
  PVRSRV_ERROR eError;
  PVRSRV_FENCE hUpdateFence;
  PVRSRV_FENCE hUpdateFence3D;
} __packed PVRSRV_BRIDGE_OUT_RGXKICKTA3D2;

struct reservation {
  uint64_t handle;
  uint64_t addr;
  uint64_t len;
  uint64_t pmr_handle;
  bool zapped;
};
struct reservation reservations[1000];
size_t num_reservations = 0;

static void *low_reservation;
static char *low_res_ptr;
static int pvrsrv_fd = -1;
#define DRM_IOCTL_PVR_SRVKM_CMD 0xc0206440

static int driver_cmd(int bridge_id, int bridge_func_id, void *in_data, void *out_data, int in_size, int out_size) {
  struct drm_pvr_srvkm_cmd cmd = {
    .bridge_id = bridge_id,
    .bridge_func_id = bridge_func_id,
    .in_data_ptr = (unsigned long)in_data,
    .out_data_ptr = (unsigned long)out_data,
    .in_data_size = in_size,
    .out_data_size = out_size
  };
  return syscall(__NR_ioctl, pvrsrv_fd, DRM_IOCTL_PVR_SRVKM_CMD, &cmd);
}
#define do_driver_cmd_noerror(bridge_id, bridge_func_id) ({                                \
    SYSCHK(driver_cmd((bridge_id), (bridge_func_id), &in, &out, sizeof(in), sizeof(out))); \
})
#define do_driver_cmd(bridge_id, bridge_func_id) ({                                        \
    do_driver_cmd_noerror((bridge_id), (bridge_func_id));                                  \
    if (out.eError != 0) errx(1, #bridge_func_id " eError %u", out.eError);                \
})

uint32_t fault_addr_base;
int ioctl(int fd, unsigned int cmd, unsigned long arg) {
  int errno_ = errno;

  if (cmd == DRM_IOCTL_PVR_SRVKM_CMD) {
    struct drm_pvr_srvkm_cmd *cmd = (void*)arg;
    void *inp_ = (void*)(unsigned long)cmd->in_data_ptr;
    void *outp_ = (void*)(unsigned long)cmd->out_data_ptr;
    printf("cmd(fd=%d, bridge_id=%u(%s), bridge_func_id=%u(%s), in_data_ptr=0x%llx, out_data_ptr=0x%llx, in_data_size=0x%x, out_data_size=0x%x)\n",
           fd, cmd->bridge_id, bridge_name(cmd->bridge_id), cmd->bridge_func_id, bridge_func_name(cmd->bridge_id, cmd->bridge_func_id),
           cmd->in_data_ptr, cmd->out_data_ptr, cmd->in_data_size, cmd->out_data_size);
    if (cmd->bridge_id == PVRSRV_BRIDGE_RGXTA3D) {
      if (cmd->bridge_func_id == PVRSRV_BRIDGE_RGXTA3D_RGXKICKTA3D2) {
        PVRSRV_BRIDGE_IN_RGXKICKTA3D2 *inp = inp_;
        PVRSRV_BRIDGE_OUT_RGXKICKTA3D2 *outp = outp_;
        if (cmd->in_data_size != sizeof(PVRSRV_BRIDGE_IN_RGXKICKTA3D2))
          errx(1, "PVRSRV_BRIDGE_IN_RGXKICKTA3D2 bad def");

        printf("  hMSAAScratchBuffer=0x%llx  hZSBuffer=0x%llx\n", inp->hMSAAScratchBuffer, inp->hZSBuffer);
        printf("  pui83DCmd=0x%llx  ui323DCmdSize=%u\n", inp->pui83DCmd, inp->ui323DCmdSize);
        printf("  pui83DPRCmd=0x%llx  ui323DPRCmdSize=%u\n", inp->pui83DPRCmd, inp->ui323DPRCmdSize);
        printf("  pui8TACmd=0x%llx  ui32TACmdSize=%u\n", inp->pui8TACmd, inp->ui32TACmdSize);
#if 1
        if (inp->ui323DCmdSize) {
          printf("  in 3D: 0x%08x  0x%08x\n", ((uint32_t*)inp->pui83DCmd)[2], ((uint32_t*)inp->pui83DCmd)[3]);
          ((uint32_t*)inp->pui83DCmd)[2] = fault_addr_base + 0x000;
          ((uint32_t*)inp->pui83DCmd)[3] = fault_addr_base + 0x100;
        }
        if (inp->ui323DPRCmdSize) {
          printf("  in 3DPR: 0x%08x  0x%08x\n", ((uint32_t*)inp->pui83DPRCmd)[2], ((uint32_t*)inp->pui83DPRCmd)[3]);
          ((uint32_t*)inp->pui83DPRCmd)[2] = fault_addr_base + 0x200;
          ((uint32_t*)inp->pui83DPRCmd)[3] = fault_addr_base + 0x300;
        }
        if (inp->ui32TACmdSize) {
          printf("  in TA: 0x%08x  0x%08x\n", ((uint32_t*)inp->pui8TACmd)[2], ((uint32_t*)inp->pui8TACmd)[3]);
          ((uint32_t*)inp->pui8TACmd)[2] = fault_addr_base + 0x400;
          ((uint32_t*)inp->pui8TACmd)[3] = fault_addr_base + 0x500;
        }
        inp->hZSBuffer = 0;
        inp->hMSAAScratchBuffer = 0;
#endif
      }
      if (cmd->bridge_func_id == PVRSRV_BRIDGE_RGXTA3D_RGXPOPULATEZSBUFFER) {
        PVRSRV_BRIDGE_IN_RGXPOPULATEZSBUFFER *inp = inp_;
        PVRSRV_BRIDGE_OUT_RGXPOPULATEZSBUFFER *outp = outp_;
        outp->eError = 0;
        outp->hsPopulation = inp->hsZSBufferKM | 0xb0000000;
        printf("  hsZSBufferKM=0x%llx\n", inp->hsZSBufferKM);
      }
    }
  }

  errno = errno_;
  int res = syscall(__NR_ioctl, fd, cmd, arg);
  errno_ = errno;

  if (cmd == DRM_IOCTL_PVR_SRVKM_CMD) {
    pvrsrv_fd = fd;

    struct drm_pvr_srvkm_cmd *cmd = (void*)arg;
    void *inp_ = (void*)(unsigned long)cmd->in_data_ptr;
    void *outp_ = (void*)(unsigned long)cmd->out_data_ptr;
    if (cmd->bridge_id == PVRSRV_BRIDGE_MM) {
      if (cmd->bridge_func_id == PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE2) {
        PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2 *inp = inp_;
        PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2 *outp = outp_;
        printf("  reserve addr=0x%llx len=0x%llx heap=0x%llx flags=0x%llx  =>  resv=0x%llx\n",
               inp->sAddress, inp->uiLength, inp->hDevmemServerHeap, inp->uiFlags, outp->hReservation);
        if (num_reservations == 1000) errx(1, "record space");
        reservations[num_reservations++] = (struct reservation) {
          .handle = outp->hReservation,
          .addr = inp->sAddress,
          .len = inp->uiLength
        };
      }
      if (cmd->bridge_func_id == PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR2) {
        PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2 *inp = inp_;
        PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2 *outp = outp_;
        printf("  map heap=0x%llx PMR=0x%llx resv=0x%llx\n",
               inp->hDevmemServerHeap, inp->hPMR, inp->hReservation);
        for (int i=0; i<num_reservations; i++) {
          if (reservations[i].handle == inp->hReservation)
            reservations[i].pmr_handle = inp->hPMR;
        }
      }
      if (cmd->bridge_func_id == PVRSRV_BRIDGE_MM_PMRLOCALIMPORTPMR) {
        PVRSRV_BRIDGE_IN_PMRLOCALIMPORTPMR *inp = inp_;
        PVRSRV_BRIDGE_OUT_PMRLOCALIMPORTPMR *outp = outp_;
        printf("  localimport ext_handle=0x%llx  =>  align=0x%llx size=0x%llx PMR=0x%llx\n",
               inp->hExtHandle, outp->uiAlign, outp->uiSize, outp->hPMR);
      }
    }
    if (cmd->bridge_id == PVRSRV_BRIDGE_RGXTA3D) {
      if (cmd->bridge_func_id == PVRSRV_BRIDGE_RGXTA3D_RGXKICKTA3D2) {
        PVRSRV_BRIDGE_IN_RGXKICKTA3D2 *inp = inp_;
        PVRSRV_BRIDGE_OUT_RGXKICKTA3D2 *outp = outp_;
        if (cmd->in_data_size != sizeof(PVRSRV_BRIDGE_IN_RGXKICKTA3D2))
          errx(1, "PVRSRV_BRIDGE_IN_RGXKICKTA3D2 bad def");

        printf("  hMSAAScratchBuffer=0x%llx  hZSBuffer=0x%llx\n", inp->hMSAAScratchBuffer, inp->hZSBuffer);
        printf("  pui83DCmd=0x%llx  ui323DCmdSize=%u\n", inp->pui83DCmd, inp->ui323DCmdSize);
        printf("  pui83DPRCmd=0x%llx  ui323DPRCmdSize=%u\n", inp->pui83DPRCmd, inp->ui323DPRCmdSize);
        printf("  pui8TACmd=0x%llx  ui32TACmdSize=%u\n", inp->pui8TACmd, inp->ui32TACmdSize);
      }
      if (cmd->bridge_func_id == PVRSRV_BRIDGE_RGXTA3D_RGXCREATEZSBUFFER2) {
        PVRSRV_BRIDGE_IN_RGXCREATEZSBUFFER2 *inp = inp_;
        PVRSRV_BRIDGE_OUT_RGXCREATEZSBUFFER2 *outp = outp_;
        printf("RGXCreateZSBuffer2(hPMR=0x%llx, hReservation=0x%llx, uiMapFlags=0x%llx) = 0x%llx\n", inp->hPMR, inp->hReservation, inp->uiMapFlags, outp->hsZSBufferKM);
      }
    }
  }

  errno = errno_;
  return res;
}

const GLchar* vertexShaderSource = "#version 320 es\n"
    "layout (location = 0) in vec3 position;\n"
    "void main()\n"
    "{\n"
    "gl_Position = vec4(position.x, position.y, position.z, 1.0);\n"
    "}\0";
const GLchar* fragmentShaderSource = "#version 320 es\n"
    "out mediump vec4 color;\n"
    "void main()\n"
    "{\n"
    "color = vec4(1.0f, 0.5f, 0.2f, 1.0f);\n"
    "}\n\0";

int main(int argc, char **argv) {
  setbuf(stdout, NULL);
  printf("pid=%d\n", getpid());

  if (argc != 2)
    errx(1, "usage: %s <fault_addr_base>", argv[0]);
  fault_addr_base = strtoul(argv[1], NULL, 0);

  EGLDisplay display = eglGetDisplay(EGL_DEFAULT_DISPLAY);
  if (display == EGL_NO_DISPLAY)
    errx(1, "no display");
  if (!eglInitialize(display, NULL, NULL))
    errx(1, "failed init");
  printf("EGL_VERSION = %s\n", eglQueryString(display, EGL_VERSION));

  EGLConfig config = NULL;
  {
    EGLConfig all_configs[10000];
    EGLint num_all_configs;
    if (!eglGetConfigs(display, all_configs, sizeof(all_configs)/sizeof(all_configs[0]), &num_all_configs))
      errx(1, "eglGetConfigs failed");
    printf("all configs: %d\n", num_all_configs);
    for (int i=0; i<num_all_configs; i++) {
      EGLConfig cfg = all_configs[i];
      EGLint red_size, green_size, blue_size, alpha_size, depth_size, level, samples, sample_buffers;
      if (!eglGetConfigAttrib(display, cfg, EGL_RED_SIZE  , &red_size  )) errx(1, "gca");
      if (!eglGetConfigAttrib(display, cfg, EGL_GREEN_SIZE, &green_size)) errx(1, "gca");
      if (!eglGetConfigAttrib(display, cfg, EGL_BLUE_SIZE , &blue_size )) errx(1, "gca");
      if (!eglGetConfigAttrib(display, cfg, EGL_ALPHA_SIZE, &alpha_size)) errx(1, "gca");
      if (!eglGetConfigAttrib(display, cfg, EGL_DEPTH_SIZE, &depth_size)) errx(1, "gca");
      if (!eglGetConfigAttrib(display, cfg, EGL_LEVEL     , &level     )) errx(1, "gca");
      if (!eglGetConfigAttrib(display, cfg, EGL_SAMPLE_BUFFERS, &sample_buffers)) errx(1, "gca");
      if (!eglGetConfigAttrib(display, cfg, EGL_SAMPLES   , &samples     )) errx(1, "gca");
      printf("red %d, green %d, blue %d, alpha %d, depth %d, level %d, samples %d, sample_buffers %d\n",
             red_size, green_size, blue_size, alpha_size, depth_size, level, samples, sample_buffers);
      if (red_size == 8 && green_size == 8 && blue_size == 8 && alpha_size == 0 && depth_size == 24 && sample_buffers == 1 && samples == 8) {
        printf("selected config\n");
        config = cfg;
      }
    }
  }
  if (config == NULL)
    errx(1, "no chosen config");

  const EGLint ctx_attrs[] = {
    EGL_CONTEXT_MAJOR_VERSION, 2,
    EGL_NONE
  };
  EGLContext ctx = eglCreateContext(display, config, NULL, ctx_attrs);
  if (ctx == EGL_NO_CONTEXT)
    errx(1, "eglCreateContext()");

  EGLint surface_attrs[] = {
    EGL_WIDTH, 1024,
    EGL_HEIGHT, 1024,
    EGL_NONE
  };
  EGLSurface my_surface = eglCreatePbufferSurface(display, config, surface_attrs);

  if (my_surface == EGL_NO_SURFACE)
    errx(1, "no surface");
  if (!eglMakeCurrent(display, my_surface, my_surface, ctx))
    fatal_egl("eglMakeCurrent");


  GLuint vertexShader = glCreateShader(GL_VERTEX_SHADER);
  glShaderSource(vertexShader, 1, &vertexShaderSource, NULL);
  glCompileShader(vertexShader);
  // Check for compile time errors
  GLint success;
  GLchar infoLog[512];
  glGetShaderiv(vertexShader, GL_COMPILE_STATUS, &success);
  if (!success) {
    char out_buf[0x10000];
    GLsizei length_out;
    glGetShaderInfoLog(vertexShader, sizeof(out_buf), &length_out, out_buf);
    printf("<<<<<\n%s>>>>>\n", out_buf);
    errx(1, "compile vertex shader");
  }

  GLuint fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
  glShaderSource(fragmentShader, 1, &fragmentShaderSource, NULL);
  glCompileShader(fragmentShader);
  // Check for compile time errors
  glGetShaderiv(fragmentShader, GL_COMPILE_STATUS, &success);
  if (!success) {
    char out_buf[0x10000];
    GLsizei length_out;
    glGetShaderInfoLog(fragmentShader, sizeof(out_buf), &length_out, out_buf);
    printf("<<<<<\n%s>>>>>\n", out_buf);
    errx(1, "compile fragment shader");
  }

  glEnable(GL_DEPTH_TEST);
  glEnable(GL_MULTISAMPLE_EXT);

  GLuint shaderProgram = glCreateProgram();
  glAttachShader(shaderProgram, vertexShader);
  glAttachShader(shaderProgram, fragmentShader);
  glLinkProgram(shaderProgram);
  // Check for linking errors
  glGetProgramiv(shaderProgram, GL_LINK_STATUS, &success);
  if (!success)
    errx(1, "link shaders");
  glDeleteShader(vertexShader);
  glDeleteShader(fragmentShader);

  GLfloat firstTriangle[] = {
      -0.9f, -0.5f, 0.0f,  // Left 
      -0.0f, -0.5f, 0.0f,  // Right
      -0.45f, 0.5f, 0.0f,  // Top 
  };
  GLfloat secondTriangle[] = {
       0.0f, -0.5f, 0.0f,  // Left
       0.9f, -0.5f, 0.0f,  // Right
       0.45f, 0.5f, 0.0f   // Top 
  };
  GLuint VBOs[2], VAOs[2];
  glGenVertexArrays(2, VAOs); // We can also generate multiple VAOs or buffers at the same time
  glGenBuffers(2, VBOs);
  // ================================
  // First Triangle setup
  // ===============================
  glBindVertexArray(VAOs[0]);
  glBindBuffer(GL_ARRAY_BUFFER, VBOs[0]);
  glBufferData(GL_ARRAY_BUFFER, sizeof(firstTriangle), firstTriangle, GL_STATIC_DRAW);
  glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(GLfloat), (GLvoid*)0); // Vertex attributes stay the same
  glEnableVertexAttribArray(0);
  glBindVertexArray(0);
  // ================================
  // Second Triangle setup
  // ===============================
  glBindVertexArray(VAOs[1]); // Note that we bind to a different VAO now
  glBindBuffer(GL_ARRAY_BUFFER, VBOs[1]); // And a different VBO
  glBufferData(GL_ARRAY_BUFFER, sizeof(secondTriangle), secondTriangle, GL_STATIC_DRAW);
  glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, (GLvoid*)0); // Because the vertex data is tightly packed we can also specify 0 as the vertex attribute's stride to let OpenGL figure it out.
  glEnableVertexAttribArray(0);
  glBindVertexArray(0);

  glClear(GL_COLOR_BUFFER_BIT);

  glUseProgram(shaderProgram);
  // Draw first triangle using the data from the first VAO
  glBindVertexArray(VAOs[0]);
  glDrawArrays(GL_TRIANGLES, 0, 3);
  // Then we draw the second triangle using the data from the second VAO
  glBindVertexArray(VAOs[1]);
  glDrawArrays(GL_TRIANGLES, 0, 3);
  glBindVertexArray(0);

  glFinish();
}
